//
// Created by tim on 21.12.23.
//

#ifndef CTUAPEX_PREPROCESSOREXTRACT_H
#define CTUAPEX_PREPROCESSOREXTRACT_H

#include <clang/Lex/PPCallbacks.h>
#include <clang/Lex/Lexer.h>
#include <spdlog/spdlog.h>

#include <utility>

#include "iostream"

#include "ExtractionBlock.h"

class PreprocessorExtract : public clang::PPCallbacks {
public:
    PreprocessorExtract(clang::SourceManager &sm, std::vector<ExtractionBlock> &includes,
                        std::vector<ExtractionBlock> &macros, std::vector<ExtractionBlock> &pragmas) : sourceManager(
            sm), includes(includes), macros(macros), pragmas(pragmas), languageOption(clang::LangOptions()) {}

    inline void InclusionDirective(clang::SourceLocation HashLoc, const clang::Token &IncludeTok,
                                   clang::StringRef FileName, bool IsAngled,
                                   clang::CharSourceRange FilenameRange,
                                   clang::Optional<clang::FileEntryRef> File,
                                   clang::StringRef SearchPath, clang::StringRef RelativePath,
                                   const clang::Module *Imported,
                                   clang::SrcMgr::CharacteristicKind FileType) override;

    void If(clang::SourceLocation Loc, clang::SourceRange ConditionRange,
            ConditionValueKind ConditionValue) override;

    void Ifdef(clang::SourceLocation Loc, const clang::Token &MacroNameTok,
               const clang::MacroDefinition &MD) override;

    void Ifndef(clang::SourceLocation Loc, const clang::Token &MacroNameTok,
               const clang::MacroDefinition &MD) override;

    void Endif(clang::SourceLocation Loc, clang::SourceLocation IfLoc) override;

    inline void PragmaDirective(clang::SourceLocation Loc,
                                clang::PragmaIntroducerKind Introducer) override;

    void Defined(const clang::Token &MacroNameTok, const clang::MacroDefinition &MD, clang::SourceRange Range) override;

private:
    std::vector<ExtractionBlock> &includes;
    std::vector<ExtractionBlock> &macros;
    std::vector<ExtractionBlock> &pragmas;
    //FixMe: make macroBeginStack with beginSourceLoc to save space
    std::stack<ExtractionBlock> macroNestingStack;
    clang::SourceManager &sourceManager;
    clang::LangOptions languageOption;

    void MacroDefined(const clang::Token &MacroNameTok, const clang::MacroDirective *MD);
};

#endif //CTUAPEX_PREPROCESSOREXTRACT_H
