//
// Created by tim on 21.12.23.
//
#include <clang/Lex/Preprocessor.h>
#include "PreprocessorExtract.h"


void PreprocessorExtract::If(clang::SourceLocation Loc, clang::SourceRange ConditionRange,
                             ConditionValueKind ConditionValue) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }
    assert(Loc.isValid());
    macroNestingStack.emplace(Loc, Loc, "");
}

void PreprocessorExtract::Ifdef(clang::SourceLocation Loc, const clang::Token &MacroNameTok,
                                const clang::MacroDefinition &MD) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }
    assert(Loc.isValid());
    macroNestingStack.emplace(Loc, Loc, "");
}

void PreprocessorExtract::Ifndef(clang::SourceLocation Loc, const clang::Token &MacroNameTok,
                                 const clang::MacroDefinition &MD) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }
    Ifdef(Loc, MacroNameTok, MD);
};

void PreprocessorExtract::Endif(clang::SourceLocation Loc, clang::SourceLocation IfLoc) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }
    assert(macroNestingStack.top().begin.isValid());
    assert(macroNestingStack.top().end.isValid());
    assert(IfLoc.isValid());
    assert("If these are not equal, we overwrote the last found macro end location already" &&
           macroNestingStack.top().begin == macroNestingStack.top().end);
    assert("The last found macros start should be the if loc we found" && macroNestingStack.top().begin == IfLoc);
    ExtractionBlock macro = macroNestingStack.top();
    macroNestingStack.pop();
    macro.end = Loc;
    clang::SourceRange sourceRange(macro.begin, macro.end);
    clang::CharSourceRange charSourceRange(sourceRange, true);
    macro.contents = "#" + clang::Lexer::getSourceText(charSourceRange, sourceManager, languageOption).str();
    macros.push_back(macro);
}

inline void PreprocessorExtract::PragmaDirective(clang::SourceLocation Loc,
                                                 clang::PragmaIntroducerKind Introducer) {
    if (!sourceManager.isInMainFile(Loc)) {
        return;
    }

    Loc.printToString(sourceManager);
    //We only get the start of the pragma
    clang::CharSourceRange charSourceRange = {{Loc, Loc}, true};
    llvm::StringRef text = clang::Lexer::getSourceText(charSourceRange, sourceManager, languageOption);

    //if the text we get from the lexer is empty, it means that the pragma is not at the location we expect it to be
    if(text.empty()){
        //I assume that this must mean it was introduced via the preprocessor itself
        //Therefore, the location should contain the Spelling location of the macro that introduced the pragma
        //This test is definitly ugly, but i didn't manage to get the spelling otherwise
        assert(Loc.printToString(sourceManager).find("<Spelling=")!=std::string::npos);
        //If this is the case, then we allready handle this pragma in the macro defintion extraction and can return safely
        //(the contents of the pragma will already be contained inside the extraction block of the parent maco
        return;
    }

    //We move forward until we encounter a newline, which finishes the pragma,
    // or enounter a \ followed by a newline, which means the pragma goes on in the next line
    for (int i = 0; !text.endswith("\n") || text.endswith("\\\n"); i++) {
        //only reset the assumed end of the sourcerange to keep object intact
        llvm::outs() << text << "\n";
        charSourceRange.setEnd(Loc.getLocWithOffset(i));
        text = clang::Lexer::getSourceText(charSourceRange, sourceManager, languageOption);
        assert(i < 5000);//This is for debugging
    }

    ExtractionBlock res;
    res.begin = Loc;
    res.end = charSourceRange.getEnd();
    res.contents = clang::Lexer::getSourceText(charSourceRange, sourceManager, languageOption);
    pragmas.push_back(res);
}

void PreprocessorExtract::InclusionDirective(clang::SourceLocation HashLoc, const clang::Token &IncludeTok,
                                             clang::StringRef FileName, bool IsAngled,
                                             clang::CharSourceRange FilenameRange,
                                             clang::Optional<clang::FileEntryRef> File,
                                             clang::StringRef SearchPath, clang::StringRef RelativePath,
                                             const clang::Module *Imported,
                                             clang::SrcMgr::CharacteristicKind FileType) {
//Note: we currently exclude headers only by name, we could also compare them according to their sourcerange
    if (!sourceManager.isInMainFile(HashLoc)) {
        return;
    }
    includes.emplace_back(HashLoc, FilenameRange.getEnd(), "#include \"" + FileName.str() + "\"");
}

void PreprocessorExtract::Defined(const clang::Token &MacroNameTok, const clang::MacroDefinition &MD,
                                  clang::SourceRange Range) {
    //This has not yet been needed, if we find a defined-clause in the main file, we have a problem
    if (sourceManager.isInMainFile(Range.getBegin())) {
        assert(false);
    }
}

void PreprocessorExtract::MacroDefined(const clang::Token &MacroNameTok, const clang::MacroDirective *MD) {
    if (!sourceManager.isInMainFile(MacroNameTok.getLocation())) {
        return;
    }
    //We are not interested in compiler defined internal macro definition
    //We can not rely on MD->getMacroInfo()->isBuiltinMacro() as it fails to trigger on built-in macros!?
    if (MacroNameTok.getLocation().printToString(sourceManager).find("<built-in>") != std::string::npos) {
        return;
    }

    ExtractionBlock res;
    res.begin = MD->getMacroInfo()->getDefinitionLoc();
    res.end = MD->getMacroInfo()->getDefinitionEndLoc();
    res.contents = clang::Lexer::getSourceText({{res.begin, res.end}, true}, sourceManager, languageOption);
    macros.push_back(res);
}